/* FILE: kl_pbc_util.h            -*-Mode: c++-*-
 * version 1.4
 * See the header file fo more...
 */

#include "vf.h"           // tensor buffer file I/O
#include "oxswarn.h"      // for messages
#include "kl_pbc_util.h"
#include "demagcoef.h"    // we use its functions CalculateSDA<ij>

/* End includes */

// #define ZERO_INF_TAILS
// We can here hard-code the infinity tails to be zero (debugging purpose only!)

// KL(m)
// Revision information, set via CVS keyword substitution
static const Oxs_WarningMessageRevisionInfo revision_info
  (__FILE__,
   "$Revision: 1.00 $",
   "$Date: 2006/01/19 09:00:00 $",
   "$Author: lebecki $",
   "Kristof M. Lebecki (lebecki@ifpan.edu.pl)");

// const REAL8m eps=1e-10;
const REAL8m eps=REAL8_EPSILON; // error used for comparisons: x==0.0 <-> fabs(x)<eps 

// Functions for calculating integration *approximation* of the *dipole* formula
// Following functions reflect integration of (appriopriate) CalculateSDA**_dip
//   function over "z" direction. For instance:
//   CalculateSDA00_aprx_** == \int CalculateSDA00_dip dz
// As this integral is definite and we have half-infinite bounadries, we need (functions):
//	_aprx_finZ relates to a finite boundary and it depends on the Z-position
//	_aprx_infZ relates to an inifinite boundary, it doesn't depend on Z
inline REALWIDE CalculateSDA00_aprx_finZ(REALWIDE x,REALWIDE y,REALWIDE z)
{ // Without the mul-factor, which is ALMOST the same as for CalculateSDA**_dip
  if(fabs(x)<eps&&fabs(y)<eps){ // i.e. x==y==0
    return z / (2.*pow3(fabs(z))); // fabs to have the sign
  } else {
    return z*(2.*pow4(x)+x*x*(y*y+z*z)-y*y*(y*y+z*z)) / \
      ( pow2(x*x+y*y)*sqrt(pow3(x*x+y*y+z*z)) );
  }
}
inline REALWIDE CalculateSDA00_aprx_infZ(REALWIDE x,REALWIDE y)
{ // Without the mul-factor, which is ALMOST the same as for CalculateSDA**_dip
  if(fabs(x)<eps&&fabs(y)<eps){ // i.e. x==y==0
    return 0;	  
  } else {
    return (x*x-y*y) / (pow2(x*x+y*y));
  }
}
// 01
inline REALWIDE CalculateSDA01_aprx_finZ(REALWIDE x,REALWIDE y,REALWIDE z)
{ // Without the mul-factor, which is ALMOST the same as for CalculateSDA**_dip
  if(fabs(x)<eps&&fabs(y)<eps){ // i.e. x==y==0
    return 0;
  } else {
    return x*y*z*(3.*(x*x+y*y)+2.*z*z) / \
      ( pow2(x*x+y*y)*sqrt(pow3(x*x+y*y+z*z)) );
  }
}
inline REALWIDE CalculateSDA01_aprx_infZ(REALWIDE x,REALWIDE y)
{ // Without the mul-factor, which is ALMOST the same as for CalculateSDA**_dip
  if(fabs(x)<eps&&fabs(y)<eps){ // i.e. x==y==0
    return 0;	  
  } else {
    return (2.*x*y) / pow2(x*x+y*y);
  }
}
// 02
inline REALWIDE CalculateSDA02_aprx_finZ(REALWIDE x,REALWIDE y,REALWIDE z)
{ // Without the mul-factor, which is ALMOST the same as for CalculateSDA**_dip
  return -x / ( sqrt(pow3(x*x+y*y+z*z)) );
}
// This function is even, so difference between +\infty and -\infty will cancell.
//   and this funtion won't be needed
//inline REALWIDE CalculateSDA02_aprx_infZ(REALWIDE x,REALWIDE y)
//{ return 0.0; }
// 22
inline REALWIDE CalculateSDA22_aprx_finZ(REALWIDE x,REALWIDE y,REALWIDE z)
{ // Without the mul-factor, which is ALMOST the same as for CalculateSDA**_dip
  return -z / ( sqrt(pow3(x*x+y*y+z*z)) );
}
// comments to avoid some compilers complaints about unused parameters
inline REALWIDE CalculateSDA22_aprx_infZ(REALWIDE /* x */,REALWIDE /* y*/)
{ return 0.0; }
// 11
inline REALWIDE CalculateSDA11_aprx_finZ(REALWIDE x,REALWIDE y,REALWIDE z)
{ return CalculateSDA00_aprx_finZ(y,x,z); }
inline REALWIDE CalculateSDA11_aprx_infZ(REALWIDE x,REALWIDE y)
{ return CalculateSDA00_aprx_infZ(y,x); }
// 12
inline REALWIDE CalculateSDA12_aprx_finZ(REALWIDE x,REALWIDE y,REALWIDE z)
{ return CalculateSDA02_aprx_finZ(y,x,z); }
// This function is even, so difference between +\infty and -\infty will cancell
//   and this funtion won't be needed
//inline REALWIDE CalculateSDA12_aprx_infZ(REALWIDE x,REALWIDE y)
//{ return CalculateSDA02_aprx_infZ(y,x); }


#ifdef ZERO_INF_TAILS

REALWIDE CalculateSDA00_tails(REALWIDE /*x*/,REALWIDE /*y*/,REALWIDE /*z*/,
			REALWIDE /*zPeriod*/, UINT4m /*n_max*/)
{ return 0.0; }
REALWIDE CalculateSDA11_tails(REALWIDE /*x*/,REALWIDE /*y*/,REALWIDE /*z*/,
			REALWIDE /*zPeriod*/, UINT4m /*n_max*/)
{ return 0.0; }
REALWIDE CalculateSDA22_tails(REALWIDE /*x*/,REALWIDE /*y*/,REALWIDE /*z*/,
			REALWIDE /*zPeriod*/, UINT4m /*n_max*/)
{ return 0.0; }
REALWIDE CalculateSDA01_tails(REALWIDE /*x*/,REALWIDE /*y*/,REALWIDE /*z*/,
			REALWIDE /*zPeriod*/, UINT4m /*n_max*/)
{ return 0.0; }
REALWIDE CalculateSDA02_tails(REALWIDE /*x*/,REALWIDE /*y*/,REALWIDE /*z*/,
			REALWIDE /*zPeriod*/, UINT4m /*n_max*/)
{ return 0.0; }
REALWIDE CalculateSDA12_tails(REALWIDE /*x*/,REALWIDE /*y*/,REALWIDE /*z*/,
			REALWIDE /*zPeriod*/, UINT4m /*n_max*/)
{ return 0.0; }

REALWIDE CalculateSDA00_errorBase(REALWIDE /*x*/,REALWIDE /*y*/,REALWIDE /*z*/,
			REALWIDE /*zPeriod*/, UINT4m /*n_max*/)
{ return 0.0; }
REALWIDE CalculateSDA11_errorBase(REALWIDE /*x*/,REALWIDE /*y*/,REALWIDE /*z*/,
			REALWIDE /*zPeriod*/, UINT4m /*n_max*/)
{ return 0.0; }
REALWIDE CalculateSDA22_errorBase(REALWIDE /*x*/,REALWIDE /*y*/,REALWIDE /*z*/,
			REALWIDE /*zPeriod*/, UINT4m /*n_max*/)
{ return 0.0; }
REALWIDE CalculateSDA01_errorBase(REALWIDE /*x*/,REALWIDE /*y*/,REALWIDE /*z*/,
			REALWIDE /*zPeriod*/, UINT4m /*n_max*/)
{ return 0.0; }
REALWIDE CalculateSDA02_errorBase(REALWIDE /*x*/,REALWIDE /*y*/,REALWIDE /*z*/,
			REALWIDE /*zPeriod*/, UINT4m /*n_max*/)
{ return 0.0; }
REALWIDE CalculateSDA12_errorBase(REALWIDE /*x*/,REALWIDE /*y*/,REALWIDE /*z*/,
			REALWIDE /*zPeriod*/, UINT4m /*n_max*/)
{ return 0.0; }


#else // not(ZERO_INF_TAILS), normal calculations follow
// Functions to calculate the *approximated* tails in infinities.
// Following functions reflect summation {from -\infty to -n_max} 
//  plus {from +n_max to +\infty}. For instance:
//   CalculateSDA00_aprx_tails ==   \sum _-\infty ^-n_max CalculateSDA00_dip
//                                 +\sum _+n_max ^-\infty CalculateSDA00_dip
// Now, we will explain the *approximation* (both n1 and n2 are very large)
//   \sum _n1 ^n2 CalculateSDA00_dip \approx 1/zPeriod * \int _n1' ^n2' CalculateSDA00_dip dz
//	,where	n1' and n2' are half-zPeriod shifted.
//		1/zPeriod part is caused by the step: discrete_summation->smooth_integration:
// Further:
// \int _{-\infty} ^{-n1} f dz + \int _{n2} ^{\infty} f dz 
//   == F(-n1) - F(n2) + F(\infty) - F(-\infty)
//   , where F = \int f
// While n1 \approx n2 and "nothing more can be simplified", we can simplify terms with 
//  \infty, as functions F are either even in Z (N02, N12) or odd (N00, N11, N22, N01)
//
// First, F is odd.
//   Thus, F(\infty) - F(-\infty) == F(\infty) + F(\infty) == 2*F(\infty)
REALWIDE CalculateSDA00_tails(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max)
{ 
  const REALWIDE n_max_05=static_cast<REALWIDE>(n_max)+0.5;	
  return ( CalculateSDA00_aprx_finZ(x,y,z-n_max_05*zPeriod) \
          -CalculateSDA00_aprx_finZ(x,y,z+n_max_05*zPeriod) \
       +2.*CalculateSDA00_aprx_infZ(x,y) )/zPeriod;
}
REALWIDE CalculateSDA11_tails(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max)
{ 
  const REALWIDE n_max_05=static_cast<REALWIDE>(n_max)+0.5;	
  return ( CalculateSDA11_aprx_finZ(x,y,z-n_max_05*zPeriod) \
          -CalculateSDA11_aprx_finZ(x,y,z+n_max_05*zPeriod) \
       +2.*CalculateSDA11_aprx_infZ(x,y) )/zPeriod;
}
REALWIDE CalculateSDA22_tails(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max)
{ 
  const REALWIDE n_max_05=static_cast<REALWIDE>(n_max)+0.5;	
  return ( CalculateSDA22_aprx_finZ(x,y,z-n_max_05*zPeriod) \
          -CalculateSDA22_aprx_finZ(x,y,z+n_max_05*zPeriod) \
       +2.*CalculateSDA22_aprx_infZ(x,y) )/zPeriod;
}
REALWIDE CalculateSDA01_tails(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max)
{ 
  const REALWIDE n_max_05=static_cast<REALWIDE>(n_max)+0.5;	
  return ( CalculateSDA01_aprx_finZ(x,y,z-n_max_05*zPeriod) \
          -CalculateSDA01_aprx_finZ(x,y,z+n_max_05*zPeriod) \
       +2.*CalculateSDA01_aprx_infZ(x,y) )/zPeriod;
}
// Now, F is even.
//   Thus, F(\infty) - F(-\infty) == F(\infty) - F(\infty) == 0.0
REALWIDE CalculateSDA02_tails(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max)
{ 
  const REALWIDE n_max_05=static_cast<REALWIDE>(n_max)+0.5;	
  return ( CalculateSDA02_aprx_finZ(x,y,z-n_max_05*zPeriod) \
          -CalculateSDA02_aprx_finZ(x,y,z+n_max_05*zPeriod) )/zPeriod;          
}
REALWIDE CalculateSDA12_tails(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max)
{ 
  const REALWIDE n_max_05=static_cast<REALWIDE>(n_max)+0.5;	
  return ( CalculateSDA12_aprx_finZ(x,y,z-n_max_05*zPeriod) \
          -CalculateSDA12_aprx_finZ(x,y,z+n_max_05*zPeriod) )/zPeriod;          
}

// The above *approximated* functions return results with given known error.
// Following functions are a base which (to know the absolute error value) must be
//  multiplied by the required relative-error value (now I think it should be
//    err_ratio_cont_to_dip*error_img_dip_**diag)
// As this error depends directly on CalculateSDA00_aprx_** values, to print this error
//  we step-down a little bit comaper to the formulas for CalculateSDA**_tails
//  (attention is paid to the sign).
REALWIDE CalculateSDA00_errorBase(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max)
{ 
  const REALWIDE n_max_05=static_cast<REALWIDE>(n_max)+0.5;	
  return ( fabs( CalculateSDA00_aprx_finZ(x,y,z-n_max_05*zPeriod) \
                +CalculateSDA00_aprx_infZ(x,y) ) \
          +fabs(-CalculateSDA00_aprx_finZ(x,y,z+n_max_05*zPeriod) \
                +CalculateSDA00_aprx_infZ(x,y) ) ) / zPeriod;
}
REALWIDE CalculateSDA11_errorBase(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max)
{ 
  const REALWIDE n_max_05=static_cast<REALWIDE>(n_max)+0.5;	
  return ( fabs( CalculateSDA11_aprx_finZ(x,y,z-n_max_05*zPeriod) \
                +CalculateSDA11_aprx_infZ(x,y) ) \
          +fabs(-CalculateSDA11_aprx_finZ(x,y,z+n_max_05*zPeriod) \
                +CalculateSDA11_aprx_infZ(x,y) ) ) / zPeriod;
}
REALWIDE CalculateSDA22_errorBase(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max)
{ 
  const REALWIDE n_max_05=static_cast<REALWIDE>(n_max)+0.5;	
  return ( fabs( CalculateSDA22_aprx_finZ(x,y,z-n_max_05*zPeriod) \
                +CalculateSDA22_aprx_infZ(x,y) ) \
          +fabs(-CalculateSDA22_aprx_finZ(x,y,z+n_max_05*zPeriod) \
                +CalculateSDA22_aprx_infZ(x,y) ) ) / zPeriod;
}
REALWIDE CalculateSDA01_errorBase(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max)
{ 
  const REALWIDE n_max_05=static_cast<REALWIDE>(n_max)+0.5;	
  return ( fabs( CalculateSDA01_aprx_finZ(x,y,z-n_max_05*zPeriod) \
                +CalculateSDA01_aprx_infZ(x,y) ) \
          +fabs(-CalculateSDA01_aprx_finZ(x,y,z+n_max_05*zPeriod) \
                +CalculateSDA01_aprx_infZ(x,y) ) ) / zPeriod;
}
// Now, F is even.
REALWIDE CalculateSDA02_errorBase(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max)
{ 
  const REALWIDE n_max_05=static_cast<REALWIDE>(n_max)+0.5;	
  return ( fabs(CalculateSDA02_aprx_finZ(x,y,z-n_max_05*zPeriod)) \
          +fabs(CalculateSDA02_aprx_finZ(x,y,z+n_max_05*zPeriod)) )/zPeriod;          
}
REALWIDE CalculateSDA12_errorBase(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max)
{ 
  const REALWIDE n_max_05=static_cast<REALWIDE>(n_max)+0.5;	
  return ( fabs(CalculateSDA12_aprx_finZ(x,y,z-n_max_05*zPeriod)) \
          +fabs(CalculateSDA12_aprx_finZ(x,y,z+n_max_05*zPeriod)) )/zPeriod;          
}
#endif // ZERO_INF_TAILS

// Determinant of a symmetric tensor 
REAL8m determinant( \
  	REALWIDE nXX, REALWIDE nYY, REALWIDE nZZ, 
  	REALWIDE nXY, REALWIDE nXZ, REALWIDE nYZ)
{  	
  return (-(pow(nXZ,2)*nYY) + 2.*nXY*nXZ*nYZ - nXX*pow(nYZ,2) - pow(nXY,2)*nZZ \
         + nXX*nYY*nZZ );
}   

// Assuming a Nij element of the tensor N is known with an absolute error dNij
//  how would it influence the vector H=N*m (for any vector m)?
// Answer: *maximum* relative error of the vector H: dH/H is given by *following* functions
//  multiplied by dNij.
// Example for N_01: dH/H = dN_01*rel_err_factor_01
// Remark#1: this analysis is valid for off-diagonal Nij elements only.
//  For diagonal element the relation is very simple: dH/H = dNii/Nii
// Remark#2: requirement: det(N)!=0
REAL8m rel_err_factor_01( \
  	REALWIDE nXX, REALWIDE nYY, REALWIDE nZZ, 
  	REALWIDE nXY, REALWIDE nXZ, REALWIDE nYZ, REAL8m det)
{ // in following code, for speed-up, you could exchange the calls pow(x,N) into powN(x)
  REAL8m factor1 = \
      pow(nXZ,4) + pow(nXX,2)*pow(nYZ,2) + pow(nXY,2)*pow(nYZ,2) + pow(nYZ,4) \
    - 2.*nYY*pow(nYZ,2)*nZZ + pow(nXX,2)*pow(nZZ,2) + 2.*pow(nXY,2)*pow(nZZ,2) \
    + pow(nYY,2)*pow(nZZ,2) - 2.*nXY*nXZ*nYZ*(nXX + nYY + 2.*nZZ) \
    + pow(nXZ,2)*(pow(nXY,2) + pow(nYY,2) + 2.*pow(nYZ,2) - 2.*nXX*nZZ);
  REAL8m factor2 = 
    sqrt(  pow(pow(nXZ,4) + 2.*nXY*nXZ*(-nXX + nYY)*nYZ \
         + (nXX - nYZ)*pow(nYZ,2)*(nXX + nYZ) \
         + pow(nXY,2)*(nXZ - nYZ)*(nXZ + nYZ) + 2.*nYY*pow(nYZ,2)*nZZ \
         + (nXX - nYY)*(nXX + nYY)*pow(nZZ,2) - pow(nXZ,2)*(pow(nYY,2) + 2.*nXX*nZZ),2) \
         + 4.*pow(pow(nXY,2)*nXZ*nYZ + nXZ*nYZ*(-pow(nXZ,2) + nXX*nYY - pow(nYZ,2) \
         + (nXX + nYY)*nZZ) + nXY*(-(nXX*pow(nYZ,2)) + pow(nXZ,2)*(-nYY + nZZ) \
         + nZZ*(pow(nYZ,2) - (nXX + nYY)*nZZ)),2));
  REAL8m res1 = sqrt( fabs(factor1-factor2)/2. ) / fabs(det);
  REAL8m res2 = sqrt( fabs(factor1+factor2)/2. ) / fabs(det); 
  // we are interested only in the higher value
  if(res1>res2) return res1;
  return res2;
}	  	 

// Helper function
UINT4m NextPowerOfTwo(UINT4m n)
{ // Returns first power of two >= n
  UINT4m m=1;
  while(m<n) m*=2;
  return m;
}

// Some constants used in the class
const String title_1[exec_order_MAX]=   \
  {"Demagnetization tensor diagonal elements: Nxx, Nyy, Nzz",     // _SimpleDemag
   "Demagnetization tensor elements: Nxx, Nxy, Nxz"};		  // _Demag
const String title_2[exec_order_MAX]=\
  {"Demagnetization tensor off-diagonal elements: Nxy, Nxz, Nyz", // _SimpleDemag
   "Demagnetization tensor elements: Nyy, Nyz, Nzz"};		  // _Demag
      
      
const UINT4m progress_steps = 100;  // how many steps
          
// DemagPBC class

// Constructor
DemagPBC::DemagPBC( 
      const Oxs_Mesh*	genmesh_param,
      const REAL4m 	zPeriod_param,
      const REAL4m 	max_celldistance_img_diag_param,
      const REAL4m 	max_celldistance_img_OFFdiag_param,
      const REAL4m 	error_img_dip_diag_param,
      const REAL4m 	error_img_dip_OFFdiag_param,
      const REAL4m 	err_ratio_cont_to_dip_param,  
      const UINT4m 	max_no_of_sum_elements_param,
      const INT4m  	max_no_of_sum_elements_debug_param,
      const String 	tensor_file_name_param,
      const String 	tensor_file_suffix_param,
      const String 	progress_script_param,
      const T_exec_order exec_order_param) 
      : genmesh(NULL),mesh(NULL),zPeriod(0.0),
        max_celldistance_img_diag	(max_celldistance_img_diag_param),
        max_celldistance_img_OFFdiag	(max_celldistance_img_OFFdiag_param),
        error_img_dip_diag		(error_img_dip_diag_param),
        error_img_dip_OFFdiag		(error_img_dip_OFFdiag_param),
        err_ratio_cont_to_dip		(err_ratio_cont_to_dip_param),
        max_no_of_sum_elements		(max_no_of_sum_elements_param),
        max_no_of_sum_elements_debug	(max_no_of_sum_elements_debug_param),
        tensor_file_suffix		(tensor_file_suffix_param),
        exec_order 			(exec_order_param),
        xdim(0),ydim(0),zdim(0),xydim(0),totalsize(0),
        dx(0.0),dy(0.0),dz(0.0),
        tensor_rw(NOTHING),tensor_file_prefix(""),important_parameters(""),
        progress_script			(progress_script_param),
        interp(NULL),
        show_progress(0),progress_total(0),progress_cnt(0),
        // n_max and other tables are initialized in the constructor body below
        A00_err_inf_tail(NULL),A11_err_inf_tail(NULL),A22_err_inf_tail(NULL),
        A01_err_inf_tail(NULL),A02_err_inf_tail(NULL),A12_err_inf_tail(NULL),
        A00_sum_coef(NULL),A11_sum_coef(NULL),A22_sum_coef(NULL),
        A01_sum_coef(NULL),A02_sum_coef(NULL),A12_sum_coef(NULL),
        tensor_buffer(NULL)
{ 
  // members that could not be initialized in the initializer list	
  n_max[n00]=0; 
  n_max[n11]=0; 
  n_max[n22]=0; 
  n_max[n01]=0; 
  n_max[n02]=0; 
  n_max[n12]=0; 
  max_apprx_err[n00]=0.0; 
  max_apprx_err[n11]=0.0; 
  max_apprx_err[n22]=0.0; 
  max_apprx_err[n01]=0.0; 
  max_apprx_err[n02]=0.0; 
  max_apprx_err[n12]=0.0; 
  sum_apprx_err[n00]=0.0; 
  sum_apprx_err[n11]=0.0; 
  sum_apprx_err[n22]=0.0; 
  sum_apprx_err[n01]=0.0; 
  sum_apprx_err[n02]=0.0; 
  sum_apprx_err[n12]=0.0; 
  num_apprx_err[n00]=0; 
  num_apprx_err[n11]=0; 
  num_apprx_err[n22]=0; 
  num_apprx_err[n01]=0; 
  num_apprx_err[n02]=0; 
  num_apprx_err[n12]=0; 
  max_tot_err[n00]=0.0; 
  max_tot_err[n11]=0.0; 
  max_tot_err[n22]=0.0; 
  max_tot_err[n01]=0.0; 
  max_tot_err[n02]=0.0; 
  max_tot_err[n12]=0.0; 
  sum_tot_err[n00]=0.0; 
  sum_tot_err[n11]=0.0; 
  sum_tot_err[n22]=0.0; 
  sum_tot_err[n01]=0.0; 
  sum_tot_err[n02]=0.0; 
  sum_tot_err[n12]=0.0; 
  num_tot_err[n00]=0; 
  num_tot_err[n11]=0; 
  num_tot_err[n22]=0; 
  num_tot_err[n01]=0; 
  num_tot_err[n02]=0; 
  num_tot_err[n12]=0; 
  
	
  // copying parameters
  
// ??? Is it normal that class memebers "const" can be changed inside the constructor?
// This only concerns pointers to other classes...

  // genmesh
  genmesh = genmesh_param;
  mesh = dynamic_cast<const Oxs_RectangularMesh*>(genmesh);
  if(mesh==NULL) {
    String msg = String("Object ")
      + String(genmesh->InstanceName())
      + String(" is not a rectangular mesh.");
    throw Oxs_Ext::Error(msg.c_str());
  }
  
  // Dimension variables dependent on the mesh
  xdim = mesh->DimX();
  ydim = mesh->DimY();
  zdim = mesh->DimZ();
  xydim=xdim*ydim;
  totalsize=xdim*ydim*zdim;
  //
  dx = mesh->EdgeLengthX();
  dy = mesh->EdgeLengthY();
  dz = mesh->EdgeLengthZ();
  // For demag calculation, all that matters is the relative
  // sizes of dx, dy and dz.  To help insure we don't run
  // outside floating point range, rescale these values so
  // largest is 1.0
  REALWIDE maxedge=dx;
  if(dy>maxedge) maxedge=dy;
  if(dz>maxedge) maxedge=dz;
  dx/=maxedge; dy/=maxedge; dz/=maxedge;
  
  // zPeriod
  // It depends on dimension variables!
  if(zPeriod_param==0) { // we take zdim in that case (atlas length)
    zPeriod = static_cast<REAL8m>(zdim) * dz;
  } else { // we take parameter value in that case
    zPeriod = static_cast<REAL8m>(zPeriod_param / (mesh->EdgeLengthZ())) * dz;
  }
  if(zPeriod_param<(static_cast<REAL8m>(zdim) * (mesh->EdgeLengthZ())) && 
     zPeriod_param!=0.0) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf), "Strange parameter value:"
  	    " Specified \"zPeriod\" is %g (should be zero or >= %g, i.e. the z-length)\n"
	      "I cannot imagine why you have done it, but as you wish...\n"
  	    ,zPeriod_param, static_cast<REAL8m>(zdim) * (mesh->EdgeLengthZ())
	  );
    static Oxs_WarningMessage foo(1);
    foo.Send(revision_info,OC_STRINGIFY(__LINE__), buf);	  
  }
  
  // tensor_file_name
  // tensor_file_name is not needed at global level
  // Instead we set up here other variables
  if(tensor_file_name_param=="")
    tensor_rw = NOTHING;
  else
    tensor_rw = WRITE;
  // Full file prefix
  if(tensor_file_name_param!="") {
    if(tensor_file_name_param.substr( tensor_file_name_param.length()-1,
	      	                      tensor_file_name_param.length()-1)=="/") {
      // it is a directory, we create a file name basing on it
      char buf[1024];
      Oc_Snprintf(buf,sizeof(buf),"t-%d-%d-%d-%d-",xdim,ydim,zdim,
      				  static_cast<UINT4m>(OC_ROUND(zPeriod/dz)));
      tensor_file_prefix = tensor_file_name_param + buf; 	      	                      
    } else
      // it is just a file name
      tensor_file_prefix = tensor_file_name_param;
    // important input parameters
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),". zP=%g,mcid=%g,mciO=%g,eidd=%g,eidO=%g,erctd=%g",
    				     zPeriod_param,
    				     	     max_celldistance_img_diag_param,
    				     	     	     max_celldistance_img_OFFdiag_param,
    				     	     	     	     error_img_dip_diag_param,
    				     	     	     	     	     error_img_dip_OFFdiag_param,
    				     	     	     	     	     	      err_ratio_cont_to_dip_param
        	);
    important_parameters = buf;
  }
  
  // progress_script
  if(progress_script_param!="") show_progress=1;
  else show_progress=0;
}  

void DemagPBC::Init() const
{ 
  // Clean-up from previous allocation, if any.
  ReleaseMemory();
  
  // maximal numbers of repetitions after which the difference apprx-dip should
  //  be below the "err_ratio_cont_to_dip*error_img_dip_APPRIOPRIATE" level
  n_max[n00] = static_cast<INT4m>(ceil( \
  	  sqrt(	zPeriod*zPeriod/(2.*err_ratio_cont_to_dip*error_img_dip_diag) \
    		   +4.*xdim*xdim )/zPeriod ));
  n_max[n01] = static_cast<INT4m>(ceil( \
  	  sqrt( 5./(4.*err_ratio_cont_to_dip*error_img_dip_OFFdiag) ) ));
  n_max[n02] = static_cast<INT4m>(ceil( \
  	  sqrt( 5./(6.*err_ratio_cont_to_dip*error_img_dip_OFFdiag) ) ));
  // The following will be quite similar to n_max[n00], but maybe someday we 
  //  will use their difference.
  n_max[n22] = static_cast<INT4m>(ceil( \
  	  sqrt(	zPeriod*zPeriod/(2.*err_ratio_cont_to_dip*error_img_dip_diag) \
    		   +2.*xdim*xdim )/zPeriod ));
  n_max[n11] = static_cast<INT4m>(ceil( \
  	  sqrt(	zPeriod*zPeriod/(2.*err_ratio_cont_to_dip*error_img_dip_diag) \
    		   +4.*ydim*ydim )/zPeriod ));
  n_max[n12] = n_max[n02]; // it does not depend on xyz-dim
  // unless we want manually set them ...
  if(max_no_of_sum_elements_debug>-1) {
    n_max[n00] = max_no_of_sum_elements_debug;
    n_max[n01] = max_no_of_sum_elements_debug;
    n_max[n02] = max_no_of_sum_elements_debug;
    n_max[n22] = max_no_of_sum_elements_debug;
    n_max[n11] = max_no_of_sum_elements_debug;
    n_max[n12] = max_no_of_sum_elements_debug;
  }
  // max_no_of_sum_elements check
  if(n_max[n00]>max_no_of_sum_elements||n_max[n01]>max_no_of_sum_elements|| \
     n_max[n02]>max_no_of_sum_elements||n_max[n11]>max_no_of_sum_elements|| \
     n_max[n22]>max_no_of_sum_elements) {
    char buf[4096];
    Oc_Snprintf(buf,sizeof(buf),
	    	"The predicted approximation-dipole switch will happen \n"
	    	"for repetitions (%d, %d, %d, %d, %d) which are larger \n"
	    	"than the \"max_no_of_sum_elements\" limit: %d\n",
	    	n_max[n00],n_max[n11],n_max[n22],n_max[n01],n_max[n02], \
	    	max_no_of_sum_elements);
    throw Oxs_Ext::Error(buf);
  }
  
  /* probably this part can be removed, as it is placed anyhow in the constructor? 
     or rather it should be removed _from_ the constructor... */
  // error introduced by the *approximation*
  max_apprx_err[n00] = 0.0; // Relative error
  sum_apprx_err[n00] = 0.0; // Sum, used later for averaging
  num_apprx_err[n00] = 0;   // Used for enumarating (to average)
  max_apprx_err[n11] = 0.0;
  sum_apprx_err[n11] = 0.0;
  num_apprx_err[n11] = 0;
  max_apprx_err[n22] = 0.0;
  sum_apprx_err[n22] = 0.0;
  num_apprx_err[n22] = 0;
  max_apprx_err[n01] = 0.0;
  sum_apprx_err[n01] = 0.0;
  num_apprx_err[n01] = 0;
  max_apprx_err[n02] = 0.0;
  sum_apprx_err[n02] = 0.0;
  num_apprx_err[n02] = 0;
  max_apprx_err[n12] = 0.0;
  sum_apprx_err[n12] = 0.0;
  num_apprx_err[n12] = 0;
  // Relative errors introduced by the *approximation*, and also by *dip* and *img*
  max_tot_err[n00] = 0.0; // Maximal value
  sum_tot_err[n00] = 0.0; // Sum, used later for averaging
  num_tot_err[n00] = 0;   // Used for enumarating (to average)
  max_tot_err[n11] = 0.0;
  sum_tot_err[n11] = 0.0;
  num_tot_err[n11] = 0;
  max_tot_err[n22] = 0.0;
  sum_tot_err[n22] = 0.0;
  num_tot_err[n22] = 0;
  max_tot_err[n01] = 0.0;
  sum_tot_err[n01] = 0.0;
  num_tot_err[n01] = 0;
  max_tot_err[n02] = 0.0;
  sum_tot_err[n02] = 0.0;
  num_tot_err[n02] = 0;
  max_tot_err[n12] = 0.0;
  sum_tot_err[n12] = 0.0;
  num_tot_err[n12] = 0;
  
  // Tensor file storage initialization
  if(tensor_rw==WRITE) {
    // In this case we will need the tensor_buffer
    tensor_buffer.AdjustSize(mesh);
    if(tensor_buffer.Size()!=mesh->Size())
      throw Oxs_Ext::Error("Error allocating space for the tensor_buffer");
    // try to read it
    if(1==load_tensor_file(tensor_file_prefix+"2"+tensor_file_suffix, 
                           title_2[exec_order]+important_parameters)) {
      if(1==load_tensor_file(tensor_file_prefix+"1"+tensor_file_suffix,
                             title_1[exec_order]+important_parameters)) {
        // we can use the old file (it is valid).
        tensor_rw = READ;
        // BTW, we have already loaded the tensor with diagonal elements
      }
    }
  }
  
  // script initialization
  if(show_progress==1 && tensor_rw!=READ) {
  }
  
  
  // Progress initialization
  // variables, constants
  progress_cnt    = 0;
  progress_total  = n_max[n00]+n_max[n11]+n_max[n22]+n_max[n01]+n_max[n02]+n_max[n12];
  // script initialization
  if(show_progress==1 && tensor_rw!=READ) {
    // in case of READ we do not progress anything
    interp=Oc_GlobalInterpreter();
    if(interp==NULL)
      throw Oxs_Ext::Error("Tcl interpretor not initialized");
    int errcode = TCL_OK;
    // now, we start the progress-script
    errcode = Tcl_Eval(interp,("set klmchan [open \"|wish "
                              +progress_script+"\" w+]").c_str());
    if(errcode==TCL_OK) {
      // Configure process channel, and double-check launch success
      errcode = Tcl_Eval(interp,"\n\
        fconfigure $klmchan -blocking 0 -buffering line\n\
        set kl_pbc_check_loop 10\n\
        while {[set kl_pbc_check_count [gets $klmchan kl_pbc_check_value]]<1} {\n\
           if {$kl_pbc_check_count<0 && [eof $klmchan]} {\n\
              fconfigure $klmchan -blocking 1\n\
              close $klmchan\n\
              error {Error launching progress script}\n\
           }\n\
           if {[incr kl_pbc_check_loop -1]<0} {\n\
              catch {close $klmchan}\n\
              error {Timeout launching progress script}\n\
           }\n\
           after 1000\n\
        }\n\
        if {[string compare \"kl_progress_OK\" $kl_pbc_check_value]!=0} {\n\
           catch {close $klmchan}\n\
           error \"Bad progress script check string: -->$kl_pbc_check_value<--\"\n\
        }\n\
      ");
    }
    if(errcode!=TCL_OK) {
      static Oxs_WarningMessage foo(-1);
      foo.Send(revision_info,OC_STRINGIFY(__LINE__),
               (String("Failure launching progress script \"")
                + progress_script + String("\": ")
                +Tcl_GetStringResult(interp)).c_str());
      show_progress=0;
    }
  }
                                     
  if(tensor_rw==WRITE || tensor_rw==NOTHING) {
    // in this case we will make the error statistics
    A00_err_inf_tail.AdjustSize(mesh);
    A11_err_inf_tail.AdjustSize(mesh);
    A22_err_inf_tail.AdjustSize(mesh);
    A01_err_inf_tail.AdjustSize(mesh);
    A02_err_inf_tail.AdjustSize(mesh);
    A12_err_inf_tail.AdjustSize(mesh);
    A00_sum_coef.AdjustSize(mesh);
    A11_sum_coef.AdjustSize(mesh);
    A22_sum_coef.AdjustSize(mesh);
    A01_sum_coef.AdjustSize(mesh);
    A02_sum_coef.AdjustSize(mesh);
    A12_sum_coef.AdjustSize(mesh);
    if(A00_err_inf_tail.Size()!=mesh->Size()||A00_sum_coef.Size()!=mesh->Size()||
       A11_err_inf_tail.Size()!=mesh->Size()||A11_sum_coef.Size()!=mesh->Size()||
       A22_err_inf_tail.Size()!=mesh->Size()||A22_sum_coef.Size()!=mesh->Size()||
       A01_err_inf_tail.Size()!=mesh->Size()||A01_sum_coef.Size()!=mesh->Size()||
       A02_err_inf_tail.Size()!=mesh->Size()||A02_sum_coef.Size()!=mesh->Size()||
       A12_err_inf_tail.Size()!=mesh->Size()||A12_sum_coef.Size()!=mesh->Size())
      throw Oxs_Ext::Error(
          "Error allocating space for the error statistics objects");
  }
}

// Dumps the tensor_buffer into a file
void DemagPBC::save_tensor_file( const String file_name,
                                 const String title,
                                 const char* desc) const
{
  const BOOL write_headers = 1;
  const char* valueunit = "{}";
  const char* meshtype = genmesh->NaturalOutputType();
//  const char* datatype  = "binary";
//  const char* precision = "8";
  const char* datatype  = "text";
  const char* precision = "%.20g";
  //
  genmesh->WriteOvf(file_name.c_str(),
     	  write_headers,
		    title.c_str(),desc,
		    valueunit,meshtype,datatype,precision,
		    &tensor_buffer,NULL);
}

// Fills the tensor_buffer from a file
BOOL DemagPBC::load_tensor_file( const String file_name, 
                                 const String title) const
// Result: succesfull (1) or not (0) load process
// For a success the file_mesh has to be *exactly* the same as the current one
// Title must be appriopriate, too.
// As in demag evaluation only relative size is important, in future one can 
// code all necessary shifting (offsets) and re-scaling to support *similar* meshes.
// As for now it is not implemented.
{
  Vf_Mesh* file_mesh;
  // Create and load Vf_Mesh object with file data
  // This line produces a message when the file is not present
  // But this message goes only to stdout? So, it should not be 
  //  a great problem. If it were, well either check for the file
  //  presence first, or what else?
  Vf_FileInput* vffi=Vf_FileInput::NewReader(file_name.c_str());
  if(vffi==NULL) 
    return 0; // probably no file present
  try { // we must ensure Delete of vfii
    file_mesh = vffi->NewMesh();
  } catch(...) {
    delete vffi;
    return 0;
  }
  delete vffi;
  
  if(file_mesh==NULL) 
    return 0; //Unable to create Vf_Mesh object on input file

  try { // we must ensure Delete of file_mesh 
    // title
    if(String(file_mesh->GetTitle())!=title) {
      char buf[4096];
      Oc_Snprintf(buf,sizeof(buf),
    		"\nInappriopriate title:\n"
    		"%s\n"
    		"%s\n"
    		,String(file_mesh->GetTitle()).c_str()
        ,title.c_str());
      throw Oxs_Ext::Error(buf);
    }
    // mesh size
    if(static_cast<UINT4m>(file_mesh->GetSize())!=mesh->Size()) {
      char buf[4096];
      Oc_Snprintf(buf,sizeof(buf),
    		"\nDifferent mesh sizes:\n"
    		"%d != %d\n"
    		,static_cast<UINT4m>(file_mesh->GetSize())
        ,mesh->Size());
      throw Oxs_Ext::Error(buf);
    }
    if(tensor_buffer.Size() != mesh->Size())
      tensor_buffer.AdjustSize(mesh); // safety, should never happen
      
    // Cellsize
    Nb_Vec3<REAL4> file_dim= file_mesh->GetApproximateCellDimensions();   
    const REAL4 edgeX = static_cast<REAL4>(mesh->EdgeLengthX());
    const REAL4 edgeY = static_cast<REAL4>(mesh->EdgeLengthY());
    const REAL4 edgeZ = static_cast<REAL4>(mesh->EdgeLengthZ());    
    if( fabs(edgeX - file_dim.x)>REAL4_EPSILON*edgeX ||
        fabs(edgeY - file_dim.y)>REAL4_EPSILON*edgeY ||
        fabs(edgeZ - file_dim.z)>REAL4_EPSILON*edgeZ ) {
      char buf[4096];
      Oc_Snprintf(buf,sizeof(buf),
    		"\nDifferent cell sizes:\n"
    		"x: %g, %g\n"
    		"y: %g, %g\n"
    		"z: %g, %g\n"
    		,edgeX, file_dim.x
    		,edgeY, file_dim.y
    		,edgeZ, file_dim.z);
      throw Oxs_Ext::Error(buf);
    }
      
    // mesh boundaries
    // file
    Nb_BoundingBox<REAL8m> file_range;
    file_mesh->GetPreciseRange(file_range); // including boundary
    REAL8m fxMin, fyMin, fzMin, fxMax, fyMax, fzMax;
    file_range.GetMinPt(fxMin, fyMin, fzMin);
    file_range.GetMaxPt(fxMax, fyMax, fzMax);
    // actual
    Oxs_Box bbox;    
    mesh->GetBoundingBox(bbox);
    // check
    if( fabs(fxMin - bbox.GetMinX())>REAL8_EPSILON*fabs(fxMin) ||
        fabs(fyMin - bbox.GetMinY())>REAL8_EPSILON*fabs(fyMin) ||
        fabs(fzMin - bbox.GetMinZ())>REAL8_EPSILON*fabs(fzMin) ||
        fabs(fxMax - bbox.GetMaxX())>REAL8_EPSILON*fabs(fxMax) ||
        fabs(fyMax - bbox.GetMaxY())>REAL8_EPSILON*fabs(fyMax) ||
        fabs(fzMax - bbox.GetMaxZ())>REAL8_EPSILON*fabs(fzMax) ) {
      char buf[4096];
      Oc_Snprintf(buf,sizeof(buf),
    		"\nDifferent mesh boundaries:\n"
    		"x_min: %g, %g\n"
    		"y_min: %g, %g\n"
    		"z_min: %g, %g\n"
    		"x_max: %g, %g\n"
    		"y_max: %g, %g\n"
    		"z_max: %g, %g\n"
    		,fxMin, bbox.GetMinX()
    		,fyMin, bbox.GetMinY()
    		,fzMin, bbox.GetMinZ()
    		,fxMax, bbox.GetMaxX()
    		,fyMax, bbox.GetMaxY()
    		,fzMax, bbox.GetMaxZ() );
      throw Oxs_Ext::Error(buf);
    }
    
    // number of cells in every direction
    // This checking is redundant, but if you are writing re-scaling code
    // you can use it...
    /* const UINT4m f_noX = \
        static_cast<UINT4m>(OC_ROUND(file_range.GetWidth() /file_dim.x));
    const UINT4m f_noY = \
        static_cast<UINT4m>(OC_ROUND(file_range.GetHeight()/file_dim.y));
    const UINT4m f_noZ = \
        static_cast<UINT4m>(OC_ROUND(file_range.GetDepth() /file_dim.z));
    const UINT4m noX = \
        static_cast<UINT4m>(OC_ROUND((bbox.GetMaxX()-bbox.GetMinX())/edgeX));
    const UINT4m noY = \
        static_cast<UINT4m>(OC_ROUND((bbox.GetMaxY()-bbox.GetMinY())/edgeY));
    const UINT4m noZ = \
        static_cast<UINT4m>(OC_ROUND((bbox.GetMaxZ()-bbox.GetMinZ())/edgeZ));
    if( f_noX!=noX || f_noY!=noY || f_noZ!=noZ ) {
      char buf[4096];
      Oc_Snprintf(buf,sizeof(buf),
    		"\nDifferent number of cells in every direction:\n"
    		"x: %d!=%d (%g, %g)\n"
    		"y: %d!=%d (%g, %g)\n"
    		"z: %d!=%d (%g, %g)\n"
    		,f_noX, noX
    		,f_noY, noY
    		,f_noZ, noZ
    		,file_range.GetWidth(), static_cast<REAL8>(bbox.GetMaxX()-bbox.GetMinX())
        ,file_range.GetHeight(),static_cast<REAL8>(bbox.GetMaxY()-bbox.GetMinY())
        ,file_range.GetDepth(), static_cast<REAL8>(bbox.GetMaxZ()-bbox.GetMinZ()));
      throw Oxs_Ext::Error(buf);
    } */
    
    // Fill the buffer
    UINT4m size = tensor_buffer.Size();
    for(UINT4m i=0;i<size;i++) {
      ThreeVector location;
      mesh->Center(i,location);
      Nb_Vec3<REAL8> pos(location.x,location.y,location.z);
      Nb_LocatedVector<REAL8> lv;
      file_mesh->FindPreciseClosest(pos,lv);
      ThreeVector value(lv.value.x,lv.value.y,lv.value.z);
      tensor_buffer[i] = value * file_mesh->GetValueMultiplier();
    }
  } catch (Oxs_Ext::Error& err) {
    /* print-out the reason "while old tensor file is wrong" for test purpose */
    static Oxs_WarningMessage foo(9, NB_MSG_INFO);
    foo.Send(revision_info,OC_STRINGIFY(__LINE__),
          (String("Mesh in the tensor file is inappriopriate: ")
          +String(err)+String("\n")).c_str()); 
    /**/
    delete file_mesh;
    return 0; 
  } catch(...) {
    static Oxs_WarningMessage foo(9);
    foo.Send(revision_info,OC_STRINGIFY(__LINE__),
    	     "Some problems in the read-out of the tensor file\n");
    delete file_mesh;
    return 0;
  }
  delete file_mesh;
  return 1; // OK, successful load
}

void DemagPBC::middle_load_or_save_tensor_file() const
{
  if(tensor_rw==WRITE)
    // filling up the tensor
    // dumping the tensor
    save_tensor_file(tensor_file_prefix+"1"+tensor_file_suffix,
                     title_1[exec_order]+important_parameters, NULL);
  if(tensor_rw==READ)
    load_tensor_file(tensor_file_prefix+"2"+tensor_file_suffix,
                     title_2[exec_order]+important_parameters);
}

REAL8m DemagPBC::GetTensorElementFromBuffer(
    const UINT4m el_idx, const T_tens_el teXX) const 
{
  switch(exec_order) {
    case _SimpleDemag:
      switch(teXX) {
        case n00:
             return tensor_buffer[el_idx].x;
        case n11:
             return tensor_buffer[el_idx].y;
        case n22:
             return tensor_buffer[el_idx].z;
        case n01:
             return tensor_buffer[el_idx].x;
        case n02:
             return tensor_buffer[el_idx].y;
        case n12:
             return tensor_buffer[el_idx].z;
      }
    case _Demag:
      switch(teXX) { // now, the nXX order is different!
        case n00:
             return tensor_buffer[el_idx].x;
        case n01:
             return tensor_buffer[el_idx].y;
        case n02:
             return tensor_buffer[el_idx].z;
        case n11:
             return tensor_buffer[el_idx].x;
        case n12:
             return tensor_buffer[el_idx].y;
        case n22:
             return tensor_buffer[el_idx].z;
      }
  }
  // this should never happen, as we evaluated all possible cases...
  String msg = String("Error in function GetTensorElementFromBuffer in ")
    + String(genmesh->InstanceName());
  throw Oxs_Ext::Error(msg.c_str());
  return 0;
}
// conceptually const
void DemagPBC::PutTensorElementToBuffer(
    const REAL8m elem, const UINT4m el_idx, const T_tens_el teXX) const
{    
  switch(exec_order) {
    case _SimpleDemag:
      switch(teXX) { 
        case n00:
             tensor_buffer[el_idx].x = elem;
             break;
        case n11:
             tensor_buffer[el_idx].y = elem;
             break;
        case n22:
             tensor_buffer[el_idx].z = elem;
             break;
        case n01:
             tensor_buffer[el_idx].x = elem;
             break;
        case n02:
             tensor_buffer[el_idx].y = elem;
             break;
        case n12:
             tensor_buffer[el_idx].z = elem;
             break;
      }
      break;
    case _Demag:
      switch(teXX) { // now, the nXX order is different!
        case n00:
             tensor_buffer[el_idx].x = elem;
             break;
        case n01:
             tensor_buffer[el_idx].y = elem;
             break;
        case n02:
             tensor_buffer[el_idx].z = elem;
             break;
        case n11:
             tensor_buffer[el_idx].x = elem;
             break;
        case n12:
             tensor_buffer[el_idx].y = elem;
             break;
        case n22:
             tensor_buffer[el_idx].z = elem;
             break;
      }
  }
}

// Function to calculate the tensor element (coef in originall simpleDemag)
// Inputs are various variables like position (ijk), size (dx...), etc.
// Most thse variables are taken from the class body.
// Inf_tails and sum_coeff are written on request to appr tables 
// (with index store_inf_sum_idx)
REALWIDE DemagPBC::CalculateSDA_PBC_really( \
    const UINT4m i, const UINT4m j, const UINT4m k, \
    const T_tens_el teXX) const
{ // conceptually const
  const REALWIDE x = dx*i;
  const REALWIDE y = dy*j;
  const REALWIDE z = dz*k;
  
  // Factor from eq. (15) with additional minus sign, scale factor 
  //  for dipole-dipole approximation formula.
  // 	const REALWIDE scale_dip = (3.*dx*dy*dz)/(4.*PI);
  // Factor from eq. (19) or (28) with additional minus sign
  //	const REALWIDE scale = -1./(4*PI*dx*dy*dz);
  // As we prepare our results to be "scale multiplicative ready",
  //  it means they will be outside multiplied by scale (not scale_dip):
  //	- *img* has to be not-multiplied
  //	- *dip* has to be multiplied by scale_dip_rel = scale_dip / scale;
  const REALWIDE scale_dip_rel = -3.*pow2(dx*dy*dz);
  
  // This will be square of appriopriate max_celldistance_img_
  REALWIDE  max_dist_SQR_XXX;
  if(teXX==n00 || teXX==n11 || teXX==n22)
    // diagonal elements
    max_dist_SQR_XXX= max_celldistance_img_diag*max_celldistance_img_diag;
  else
    // OFF-diagonal elements
    max_dist_SQR_XXX= max_celldistance_img_OFFdiag*max_celldistance_img_OFFdiag;
  
  // Now, we substitute depending on teXX:
  //  appriopriate functiones ...
  REALWIDE \
    (*CalculateSDAXX)(REALWIDE,REALWIDE,REALWIDE,REALWIDE,REALWIDE,REALWIDE) = NULL,
    (*CalculateSDAXX_dip)(REALWIDE,REALWIDE,REALWIDE)                        = NULL,
    (*CalculateSDAXX_tails)(REALWIDE,REALWIDE,REALWIDE,REALWIDE,UINT4m)      = NULL,
    (*CalculateSDAXX_errorBase)(REALWIDE,REALWIDE,REALWIDE,REALWIDE,UINT4m)  = NULL;
  switch(teXX) {
    case n00: 
      CalculateSDAXX =           CalculateSDA00; 
      CalculateSDAXX_dip =       CalculateSDA00_dip; 
      CalculateSDAXX_tails =     CalculateSDA00_tails; 
      CalculateSDAXX_errorBase = CalculateSDA00_errorBase; 
      break;
    case n11: 
      CalculateSDAXX =           CalculateSDA11; 
      CalculateSDAXX_dip =       CalculateSDA11_dip; 
      CalculateSDAXX_tails =     CalculateSDA11_tails; 
      CalculateSDAXX_errorBase = CalculateSDA11_errorBase; 
      break;
    case n22: 
      CalculateSDAXX =           CalculateSDA22; 
      CalculateSDAXX_dip =       CalculateSDA22_dip; 
      CalculateSDAXX_tails =     CalculateSDA22_tails; 
      CalculateSDAXX_errorBase = CalculateSDA22_errorBase; 
      break;
    case n01: 
      CalculateSDAXX =           CalculateSDA01; 
      CalculateSDAXX_dip =       CalculateSDA01_dip; 
      CalculateSDAXX_tails =     CalculateSDA01_tails; 
      CalculateSDAXX_errorBase = CalculateSDA01_errorBase; 
      break;
    case n02: 
      CalculateSDAXX =           CalculateSDA02; 
      CalculateSDAXX_dip =       CalculateSDA02_dip; 
      CalculateSDAXX_tails =     CalculateSDA02_tails; 
      CalculateSDAXX_errorBase = CalculateSDA02_errorBase; 
      break;
    case n12: 
      CalculateSDAXX =           CalculateSDA12; 
      CalculateSDAXX_dip =       CalculateSDA12_dip; 
      CalculateSDAXX_tails =     CalculateSDA12_tails; 
      CalculateSDAXX_errorBase = CalculateSDA12_errorBase; 
      break;
  }
  
  REALWIDE coef; // result of the function
  // 0_z-repetinion
  if(x*x+y*y+z*z < max_dist_SQR_XXX)
    coef = (*CalculateSDAXX)(x,y,z,dx,dy,dz);               // not scaled
  else coef = scale_dip_rel * (*CalculateSDAXX_dip)(x,y,z); // scaled
  //
  REALWIDE x2y2=x*x+y*y; // this value will be repetitively used
  REALWIDE sum_coef=0;   // sum of the coefficients at "reapeated" cells.
  volatile REALWIDE summation_corrector = 0.0; // For compensated summation
  // n_z-repetinion (n and -n)
  // we sum from lowest "coef" values to highest
  for(UINT4m n=n_max[teXX]; n>=1; n--) {
    const REALWIDE n_zPeriod=(static_cast<REALWIDE>(n))*zPeriod;
    REALWIDE part;
    // first we evaluate z+n*zPeriod
    if(x2y2+(z+n_zPeriod)*(z+n_zPeriod) < max_dist_SQR_XXX) 
      part = (*CalculateSDAXX)(x,y,z+n_zPeriod,dx,dy,dz);           // not scaled  
    else part = scale_dip_rel * (*CalculateSDAXX_dip)(x,y,z+n_zPeriod); // scaled
    // Now, we evaluate z-n_zPeriod
    if(x2y2+(z-n_zPeriod)*(z-n_zPeriod) < max_dist_SQR_XXX) 
      part+= (*CalculateSDAXX)(x,y,z-n_zPeriod,dx,dy,dz);           // not scaled  
    else part+= scale_dip_rel * (*CalculateSDAXX_dip)(x,y,z-n_zPeriod); // scaled
    // Add up using compensated summation.
    volatile REALWIDE part_corr = part;           // this will be a summation corrector
    volatile REALWIDE sum_corr = sum_coef + part; // this will be a corrected sum
    part_corr -= sum_corr-sum_coef;         
    summation_corrector += part_corr;
    sum_coef = sum_corr;      
  }
  
  // "Infinite tails"
  const REALWIDE infinity_tails = 
  	fabs(scale_dip_rel/3.) * (*CalculateSDAXX_tails)(x,y,z,zPeriod,n_max[teXX]);
  REAL4m inf_tail_error = err_ratio_cont_to_dip* \
  	fabs(scale_dip_rel/3.) * (*CalculateSDAXX_errorBase)(x,y,z,zPeriod,n_max[teXX]);
  if(teXX==n00 || teXX==n11 || teXX==n22)
    // diagonal elements
    inf_tail_error *= error_img_dip_diag;
  else
    // OFF-diagonal elements
    inf_tail_error *= error_img_dip_OFFdiag;
    
  // We add everything together:
  coef = coef + sum_coef + summation_corrector + infinity_tails;
  
  // Error statistics
  // It can be done (here) only for diagonal elements.
  // Off-diagonals are evaluated later, basing on all-tensor-elements.
  if(teXX==n00 || teXX==n11 || teXX==n22) {
    if(coef!=0.0) {
      // Errors introduced be the infinite_tail approximation
      REAL8m tmp= fabs(inf_tail_error/coef);
      if(max_apprx_err[teXX]<tmp) {
         max_apprx_err[teXX]=tmp;}
      sum_apprx_err[teXX] += tmp*tmp;
      num_apprx_err[teXX]++;
      // Errors of the total tensor element value
      // Here, we asume that fabs(sum_coef) = \sum fabs(part)
      // This is true for far distances where the symmetricy of the tensor 
      // elements is \approx present. For highly non-symmetric cases, like in the box 
      // (1,1,1)-(n,n,n) this will be the cell (n,n,n) this will fail and we will 
      // underestimate the error. This will be seen in the trace check.
      // This could be improved if we move this error procedure inside into the PBC-loop
      tmp+= error_img_dip_diag*( fabs(coef-infinity_tails-sum_coef-summation_corrector) \
                               + fabs(sum_coef)) / fabs(coef);
      if(max_tot_err[teXX] < tmp) {
         max_tot_err[teXX] = tmp; }
      sum_tot_err[teXX] +=   tmp*tmp;
      num_tot_err[teXX]++;
    } else {
      char buf[1024];
      Oc_Snprintf(buf,sizeof(buf),"Diagonal tensor element equals to zero.\n"
        "In this case the system won't be able to calculate relative inaccuracy of "
        "the appriopriate tensor elements. The inaccuracy will be taken zero.\n"
        );
      static Oxs_WarningMessage foo(1);
      foo.Send(revision_info,OC_STRINGIFY(__LINE__), buf);
    }
  }
  
  // We store inf_tail_error, sum_coef also
  // We also update the class statistics variables
  const UINT4m store_inf_sum_idx = i+j*xdim+k*xydim; 
  switch(teXX) {
    case n00: 
      A00_err_inf_tail[store_inf_sum_idx] = inf_tail_error;
      A00_sum_coef[store_inf_sum_idx]     = sum_coef;
      break;
    case n11: 
      A11_err_inf_tail[store_inf_sum_idx] = inf_tail_error;
      A11_sum_coef[store_inf_sum_idx]     = sum_coef;
      break;
    case n22: 
      A22_err_inf_tail[store_inf_sum_idx] = inf_tail_error;
      A22_sum_coef[store_inf_sum_idx]     = sum_coef;
      break;
    case n01: 
      A01_err_inf_tail[store_inf_sum_idx] = inf_tail_error;
      A01_sum_coef[store_inf_sum_idx]     = sum_coef;
      break;
    case n02: 
      A02_err_inf_tail[store_inf_sum_idx] = inf_tail_error;
      A02_sum_coef[store_inf_sum_idx]     = sum_coef;
      break;
    case n12: 
      A12_err_inf_tail[store_inf_sum_idx] = inf_tail_error;
      A12_sum_coef[store_inf_sum_idx]     = sum_coef;
      break;
  }

  return coef;
}

REALWIDE DemagPBC::CalculateSDA(const T_tens_el teXX,
                  const UINT4m i, const UINT4m j, const UINT4m k) const
{ // conceptually const
  REALWIDE coef;

  if(tensor_rw==NOTHING || tensor_rw==WRITE) {
    // we have to calculate it ...
    if(show_progress==1) { // progress
      UINT4m nominator=0;
      switch(exec_order) {
        case _SimpleDemag:
          nominator = ((i+j*xdim+k*xydim)*n_max[teXX])/totalsize;
          // This code assumes that coefficients are
          // calculated IN TURN in the following order: 00, 11, 22, 01, 02, 12
          switch(teXX) {
            case n00: nominator += 0;
                      break;
            case n11: nominator += n_max[n00];
                      break;
            case n22: nominator += n_max[n00]+n_max[n11];
                      break;
            case n01: nominator += n_max[n00]+n_max[n11]+n_max[n22];
                      break;
            case n02: nominator += n_max[n00]+n_max[n11]+n_max[n22]+n_max[n01];
                      break;
            case n12: nominator += n_max[n00]+n_max[n11]+n_max[n22]+n_max[n01]+n_max[n02];
                      break;
          }
          break;
        case _Demag:
          // This code assumes that coefficients are
          // calculated IN TURN in the following order: 00, 11, 22, 01, 02, 12
          switch(teXX) {
            case n00: // first three are equal
              nominator = ((i+j*xdim+k*xydim)
                          *(n_max[n00]+n_max[n01]+n_max[n02]))/totalsize;
              break;
            case n01: 
              nominator = ((i+j*xdim+k*xydim)
                          *(n_max[n00]+n_max[n01]+n_max[n02]))/totalsize;
              break;
            case n02: 
              nominator = ((i+j*xdim+k*xydim)
                          *(n_max[n00]+n_max[n01]+n_max[n02]))/totalsize;
              break;
            case n11: // these three are equal, as well
              nominator = ((i+j*xdim+k*xydim)
                          *(n_max[n11]+n_max[n12]+n_max[n22]))/totalsize
                         +n_max[n00]+n_max[n01]+n_max[n02];
              break;
            case n12: 
              nominator = ((i+j*xdim+k*xydim)
                          *(n_max[n11]+n_max[n12]+n_max[n22]))/totalsize
                         +n_max[n00]+n_max[n01]+n_max[n02];
              break;
            case n22: 
              nominator = ((i+j*xdim+k*xydim)
                          *(n_max[n11]+n_max[n12]+n_max[n22]))/totalsize
                         +n_max[n00]+n_max[n01]+n_max[n02];
              break;
          }
          break;
      }
      // what can we report
      UINT4m actual_cnt = (nominator * progress_steps) / progress_total;
      if(actual_cnt > progress_cnt) {    // is it different from prev. value?
	      progress_cnt = actual_cnt;
        char buf[100];
        sprintf(buf,"puts $klmchan %d",progress_cnt);
        int errcode = TCL_OK;
        errcode = Tcl_Eval(interp,buf);
        if(errcode!=TCL_OK)
          throw Oxs_Ext::Error((String("Tcl error\n")
                                    +Tcl_GetStringResult(interp)).c_str());
      }    
    }
    coef = CalculateSDA_PBC_really(i, j, k, teXX);
        
    if(tensor_rw==WRITE)
      // filling up the tensor
      PutTensorElementToBuffer(coef, i+j*xdim+k*xydim, teXX);
  } else
    // we READ the tensor
    coef=GetTensorElementFromBuffer(i+j*xdim+k*xydim, teXX);
  
  return coef;
}                  

void DemagPBC::ReleaseMemory() const
{ // Conceptually const
/*if(A00_err_inf_tail!=NULL) 
    {A00_err_inf_tail.Release(); A00_err_inf_tail=NULL;}
  if(A11_err_inf_tail!=NULL) 
    {A11_err_inf_tail.Release(); A11_err_inf_tail=NULL;}
  if(A22_err_inf_tail!=NULL) 
    {A22_err_inf_tail.Release(); A22_err_inf_tail=NULL;}
  if(A01_err_inf_tail!=NULL) 
    {A01_err_inf_tail.Release(); A01_err_inf_tail=NULL;}
  if(A02_err_inf_tail!=NULL) 
    {A02_err_inf_tail.Release(); A02_err_inf_tail=NULL;}
  if(A12_err_inf_tail!=NULL) 
    {A12_err_inf_tail.Release(); A12_err_inf_tail=NULL;}
  //  
  if(A00_sum_coef!=NULL) 
    {A00_sum_coef.Release(); A00_sum_coef=NULL;}
  if(A11_sum_coef!=NULL) 
    {A11_sum_coef.Release(); A11_sum_coef=NULL;}
  if(A22_sum_coef!=NULL) 
    {A22_sum_coef.Release(); A22_sum_coef=NULL;}
  if(A01_sum_coef!=NULL) 
    {A01_sum_coef.Release(); A01_sum_coef=NULL;}
  if(A02_sum_coef!=NULL) 
    {A02_sum_coef.Release(); A02_sum_coef=NULL;}
  if(A12_sum_coef!=NULL) 
    {A12_sum_coef.Release(); A12_sum_coef=NULL;}
  //
  if(tensor_buffer!=NULL) 
    {tensor_buffer.Release(); tensor_buffer=NULL;} */
  A00_err_inf_tail.Release();
  A11_err_inf_tail.Release();
  A22_err_inf_tail.Release();
  A01_err_inf_tail.Release();
  A02_err_inf_tail.Release();
  A12_err_inf_tail.Release();
  A00_sum_coef.Release();
  A11_sum_coef.Release();
  A22_sum_coef.Release();
  A01_sum_coef.Release();
  A02_sum_coef.Release();
  A12_sum_coef.Release();
  tensor_buffer.Release(); 
}

void DemagPBC::Finalize(
    const Oxs_Complex *A00, const Oxs_Complex *A11, const Oxs_Complex *A22, 
    const Oxs_Complex *A01, const Oxs_Complex *A02, const Oxs_Complex *A12
) const
{
  // Statistics show buffer
  char buf_statistics[10024]=
       "statistics are available only while running Klm_SimpleDemag_PBC";
	
  // Statistics, error estimations for off-diagonal elements
  // We can do it only, if Buffers for error statistics calculation are filled / used
  // We write it to the file (only)
  if(tensor_rw==WRITE && A00!=NULL && A11!=NULL && A22!=NULL
                      && A01!=NULL && A02!=NULL && A12!=NULL ) {
    
    // Padded size will be used in Aij table access
    const UINT4m pxdim = NextPowerOfTwo(2*xdim);   
    const UINT4m pxydim= NextPowerOfTwo(2*xdim) * NextPowerOfTwo(2*ydim);
    
    // trace should be zero (except x==y==z==0), det should be non-zero 
    //  (for error evaluation)
    REAL8m max_trace = 0.0;
    REAL8m sum_trace = 0.0;
    UINT4m num_trace = 0;
    REAL8m min_det = 1e10;  // very large number...
    REAL8m sum_det = 0.0;   // I do not use it so maybe remove it someday?

    for(UINT4m k=0;k<zdim;k++) for(UINT4m j=0;j<ydim;j++) for(UINT4m i=0;i<xdim;i++) {
      // these are the tensor values
      const REAL8m c00 = A00[i+j*pxdim+k*pxydim].real();
      const REAL8m c11 = A11[i+j*pxdim+k*pxydim].real();
      const REAL8m c22 = A22[i+j*pxdim+k*pxydim].real();
      const REAL8m c01 = A01[i+j*pxdim+k*pxydim].real();
      const REAL8m c02 = A02[i+j*pxdim+k*pxydim].real();
      const REAL8m c12 = A12[i+j*pxdim+k*pxydim].real();
      // these are the tensor cont-errors (positive)
      const REAL8m i00 = A00_err_inf_tail[i+j*xdim+k*xydim];
      const REAL8m i11 = A11_err_inf_tail[i+j*xdim+k*xydim];
      const REAL8m i22 = A22_err_inf_tail[i+j*xdim+k*xydim];
      const REAL8m i01 = A01_err_inf_tail[i+j*xdim+k*xydim];
      const REAL8m i02 = A02_err_inf_tail[i+j*xdim+k*xydim];
      const REAL8m i12 = A12_err_inf_tail[i+j*xdim+k*xydim];
      // these are partial tensor values: results of finite summation of PBC-repetitions
      // (the central, 0-indexed element is not inside)
      // these (multiplied by error_img_dip_*) are the sum_coef-errors
      const REAL8m s00 = A00_sum_coef[i+j*xdim+k*xydim];
      const REAL8m s11 = A11_sum_coef[i+j*xdim+k*xydim];
      const REAL8m s22 = A22_sum_coef[i+j*xdim+k*xydim];
      const REAL8m s01 = A01_sum_coef[i+j*xdim+k*xydim];
      const REAL8m s02 = A02_sum_coef[i+j*xdim+k*xydim];
      const REAL8m s12 = A12_sum_coef[i+j*xdim+k*xydim];
      //
      REAL8m comp_err;
      // Determinant of the tensor, will be used below.
      // Is it maybe null? If yes, my algorithm fails :(
      REAL8m det = determinant(c00,c11,c22,c01,c02,c12);
      if(det==0) {
        char buf[1024];
        Oc_Snprintf(buf,sizeof(buf),"Tensor determinant equals to zero.\n"
          "In this case the system won't be able to calculate relative inaccuracy of "
          "the appriopriate tensor elements. The inaccuracy will be taken zero. "
          "Following is the list of these elements.\n"
          "xx:%g yy:%g zz:%g\n"
          "xy:%g xz:%g yz:%g\n"
          ,c00,c11,c22
          ,c01,c02,c12
          );
        static Oxs_WarningMessage foo(3);
        foo.Send(revision_info,OC_STRINGIFY(__LINE__), buf);
      } 
      else
      {
        // 01, *cont* errors
        comp_err = i01 * rel_err_factor_01(c00,c11,c22,c01,c02,c12,det);
        if( max_apprx_err[n01]<comp_err ) {
            max_apprx_err[n01]=comp_err;}
        sum_apprx_err[n01] += comp_err*comp_err;
        num_apprx_err[n01]++;
        // 01, whole element value errors
        // The error of the img+dip value is calculated basing on respectively: c-s + s
        // Well, this assumption is good only for Nxy (even in Z). Nxz and Nyz are odd!
        comp_err = ( i01 +error_img_dip_OFFdiag*(fabs(c01-s01)+fabs(s01)) ) 
                 * rel_err_factor_01(c00,c11,c22,c01,c02,c12,det);
        if( max_tot_err[n01]<comp_err ) {
            max_tot_err[n01]=comp_err;}
        sum_tot_err[n01] += comp_err*comp_err;
        num_tot_err[n01]++;
        // 02
        comp_err = i02 * rel_err_factor_02(c00,c11,c22,c01,c02,c12,det);
        if( max_apprx_err[n02]<comp_err ) {
            max_apprx_err[n02]=comp_err;}
        sum_apprx_err[n02] += comp_err*comp_err;
        num_apprx_err[n02]++;
        // 02, whole element value errors
        comp_err = ( i02+error_img_dip_OFFdiag*(fabs(c02-s02)+fabs(s02)) \
                   ) * rel_err_factor_02(c00,c11,c22,c01,c02,c12,det);
        if( max_tot_err[n02]<comp_err ) {
            max_tot_err[n02]=comp_err;}
        sum_tot_err[n02] += comp_err*comp_err;
        num_tot_err[n02]++;
        // 12
        comp_err = i12 * rel_err_factor_12(c00,c11,c22,c01,c02,c12,det);
        if( max_apprx_err[n12]<comp_err ) {
            max_apprx_err[n12]=comp_err;}
        sum_apprx_err[n12] += comp_err*comp_err;
        num_apprx_err[n12]++;
        // 12, whole element value errors
        comp_err = ( i12+error_img_dip_OFFdiag*(fabs(c12-s12)+fabs(s12)) \
                   ) * rel_err_factor_12(c00,c11,c22,c01,c02,c12,det);
        if( max_tot_err[n12]<comp_err ) {
            max_tot_err[n12]=comp_err;}
        sum_tot_err[n12] += comp_err*comp_err;
        num_tot_err[n12]++;
      }
      // statistics: determinant (should be non-zero)
      // We will normalize by dividing it by 
      //  "appriopriate" part of the relative-errors-equations.
      REAL8m det_normalized = 1 / rel_err_factor_01(c00,c11,c22,c01,c02,c12,det);
      if(det_normalized < 1.0/rel_err_factor_02(c00,c11,c22,c01,c02,c12,det))
         det_normalized = 1.0/rel_err_factor_02(c00,c11,c22,c01,c02,c12,det);
      if(det_normalized < 1.0/rel_err_factor_12(c00,c11,c22,c01,c02,c12,det)) {
         det_normalized = 1.0/rel_err_factor_12(c00,c11,c22,c01,c02,c12,det);}
      if(min_det>det_normalized)
         min_det=det_normalized;
      sum_det += det_normalized;
      // statistics: trace (should be zero).
      // We will normalize it by dividing it by the sum of diagonal (absolute) errors.
      // Again, the error of the img+dip value is calculated basing on respectively: 
      //  c-s + s
      REAL8m diag_total_abs_error= \
         i00+i11+i22+error_img_dip_diag*(fabs(c00-s00)+fabs(c11-s11)+fabs(c22-s22) \
                                        +fabs(s00)+fabs(s11)+fabs(s22));
      if(  (i!=0 || j!=0 || k!=0)    // 000 is the only case, when trace can be non-zero
      	&& diag_total_abs_error!=0 ) // safety
      { 
        comp_err = fabs( c00+c11+c22 ) / diag_total_abs_error;
        if(max_trace<comp_err)
           max_trace=comp_err;
//        sum_trace += comp_err; // arithmetic average
        sum_trace += comp_err*comp_err; // square root of powers-of-2
        num_trace++;
      }
    }
    // Information show
    Oc_Snprintf(buf_statistics,sizeof(buf_statistics),
      "PBC Information\n"
  	  "*** Parameters used ***\n"
  	  "zPeriod=%g, max_celldistance_img_diag=%g, max_celldistance_img_OFFdiag=%g,\n"
  	  "error_img_dip_diag=%g, error_img_dip_OFFdiag=%g, err_ratio_cont_to_dip=%g,\n"
  	  "max_no_of_sum_elements=%d, max_no_of_sum_elements_debug=%d\n"
  	  "*** Relative inaccuracy summary ***\n"
  	  " Averages here mean \"square root of sum of power-of-2\"\n"
  	  "Demag. tensor inaccuracy estimation (maximal values):\n"
  	  "\t%.1g\t%.1g\t(%.1g*)\n"
  	  "\t\t%.1g\t(%.1g*)\n"
  	  "\t\t\t%.1g\n"
  	  "  Their average:\n"
  	  "\t%.1g\t%.1g\t(%.1g*)\n"
  	  "\t\t%.1g\t(%.1g*)\n"
  	  "\t\t\t%.1g\n"
  	  "Inaccuracy introduced only by the \"continuous approximation\":\n"
  	  "\t%.1g\t%.1g\t%.1g\n"
  	  "\t\t%.1g\t%.1g\n"
  	  "\t\t\t%.1g\n"
  	  "  Their average:\n"
  	  "\t%.1g\t%.1g\t%.1g\n"
  	  "\t\t%.1g\t%.1g\n"
  	  "\t\t\t%.1g\n"
  	  "*- doubtful values (underestimated)\n"
  	  "*** Absolute inaccuracy check ***\n"
  	  " Trace, divided by the sum of inaccuracies\n"
  	  " (should be small, smaller than one)\n"
  	  " maximum:%g, average: %g\n"
  	  "*** Other information ***\n"
  	  " Number of positive discrete repetitions:\n"
  	  "%d %d %d %d %d %d\n"
      ,zPeriod, max_celldistance_img_diag, max_celldistance_img_OFFdiag
      ,error_img_dip_diag, error_img_dip_OFFdiag, err_ratio_cont_to_dip
      ,max_no_of_sum_elements, max_no_of_sum_elements_debug
      ,max_tot_err[n00]
      ,max_tot_err[n01]
      ,max_tot_err[n02]
      ,max_tot_err[n11]
      ,max_tot_err[n12]
      ,max_tot_err[n22]
      ,sqrt(sum_tot_err[n00])/static_cast<REAL8m>(num_tot_err[n00])
      ,sqrt(sum_tot_err[n01])/static_cast<REAL8m>(num_tot_err[n01])
      ,sqrt(sum_tot_err[n02])/static_cast<REAL8m>(num_tot_err[n02])
      ,sqrt(sum_tot_err[n11])/static_cast<REAL8m>(num_tot_err[n11])
      ,sqrt(sum_tot_err[n12])/static_cast<REAL8m>(num_tot_err[n12])
      ,sqrt(sum_tot_err[n22])/static_cast<REAL8m>(num_tot_err[n22])
      ,max_apprx_err[n00]
      ,max_apprx_err[n01]
      ,max_apprx_err[n02]
      ,max_apprx_err[n11]
      ,max_apprx_err[n12]
      ,max_apprx_err[n22]
      ,sqrt(sum_apprx_err[n00])/static_cast<REAL8m>(num_apprx_err[n00])
      ,sqrt(sum_apprx_err[n01])/static_cast<REAL8m>(num_apprx_err[n01])
      ,sqrt(sum_apprx_err[n02])/static_cast<REAL8m>(num_apprx_err[n02])
      ,sqrt(sum_apprx_err[n11])/static_cast<REAL8m>(num_apprx_err[n11])
      ,sqrt(sum_apprx_err[n12])/static_cast<REAL8m>(num_apprx_err[n12])
      ,sqrt(sum_apprx_err[n22])/static_cast<REAL8m>(num_apprx_err[n22])
      ,max_trace
      ,sqrt(sum_trace)/static_cast<REAL8m>(num_trace)
      ,n_max[n00],n_max[n11],n_max[n22],n_max[n01],n_max[n02],n_max[n12]
    );
  }
    
  // dumping the tensor
  // as desription we give here the above statistics info
  if(tensor_rw==WRITE)
    save_tensor_file(tensor_file_prefix+"2"+tensor_file_suffix, 
                     title_2[exec_order]+important_parameters, buf_statistics);
	  
  // Progress closure
  if(show_progress==1 && tensor_rw!=READ) {
    int errcode = TCL_OK;
    errcode = Tcl_Eval(interp,"puts $klmchan exit");
    if(errcode!=TCL_OK)
      throw Oxs_Ext::Error((String("Tcl error\n")
                                +Tcl_GetStringResult(interp)).c_str());
  }
  
  ReleaseMemory();
  
}

/*
if(i==3 && j==3 && k==3 && (teXX==n00 || teXX==n11 || teXX==n22) &&
(n==10000 || n==100 || n==1))         
{
        char buf[1024];
        
        Oc_Snprintf(buf,sizeof(buf),"Dupa2...\n"
	  "%d %d\n"
	  "%.20g %.20g %.20g\n"
	  "%.20g\n"
	  ,n,teXX
	  ,sum_coef, part, scale_dip_rel * (*CalculateSDAXX_dip)(x,y,z+n_zPeriod)
	  ,error_img_dip_diag * fabs(sum_coef)
	  );
//	throw Oxs_Ext::Error(buf);
        static Oxs_WarningMessage foo(9);
        foo.Send(revision_info,OC_STRINGIFY(__LINE__), buf);
}
*/
